@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2025, The PurpleI2P Project
+* Copyright (c) 2013-2026, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -11,6 +11,7 @@
 
 #include <inttypes.h>
 #include <string>
+#include <string_view>
 #include <map>
 #include <list>
 #include <set>
@@ -22,6 +23,7 @@
 #include "Identity.h"
 #include "LeaseSet.h"
 #include "Streaming.h"
+#include "Datagram.h"
 #include "Destination.h"
 
 namespace i2p
@@ -34,25 +36,25 @@ namespace client
 	const int SAM_SOCKET_CONNECTION_MAX_IDLE = 3600; // in seconds
 	const int SAM_SESSION_READINESS_CHECK_INTERVAL = 3; // in seconds
 	const size_t SAM_SESSION_MAX_ACCEPT_QUEUE_SIZE = 50;
-	const size_t SAM_SESSION_MAX_ACCEPT_INTERVAL = 3; // in seconds	
-	
+	const size_t SAM_SESSION_MAX_ACCEPT_INTERVAL = 3; // in seconds
+
 	const char SAM_HANDSHAKE[] = "HELLO VERSION";
 	const char SAM_HANDSHAKE_REPLY[] = "HELLO REPLY RESULT=OK VERSION=%s\n";
-	const char SAM_HANDSHAKE_NOVERSION[] = "HELLO REPLY RESULT=NOVERSION\n";
+	constexpr std::string_view SAM_HANDSHAKE_NOVERSION { "HELLO REPLY RESULT=NOVERSION\n" };
 	const char SAM_HANDSHAKE_I2P_ERROR[] = "HELLO REPLY RESULT=I2P_ERROR\n";
 	const char SAM_SESSION_CREATE[] = "SESSION CREATE";
 	const char SAM_SESSION_CREATE_REPLY_OK[] = "SESSION STATUS RESULT=OK DESTINATION=%s\n";
-	const char SAM_SESSION_CREATE_DUPLICATED_ID[] = "SESSION STATUS RESULT=DUPLICATED_ID\n";
-	const char SAM_SESSION_CREATE_DUPLICATED_DEST[] = "SESSION STATUS RESULT=DUPLICATED_DEST\n";
-	const char SAM_SESSION_CREATE_INVALID_ID[] = "SESSION STATUS RESULT=INVALID_ID\n";
-	const char SAM_SESSION_STATUS_INVALID_KEY[] = "SESSION STATUS RESULT=INVALID_KEY\n";
+	constexpr std::string_view SAM_SESSION_CREATE_DUPLICATED_ID { "SESSION STATUS RESULT=DUPLICATED_ID\n" };
+	constexpr std::string_view SAM_SESSION_CREATE_DUPLICATED_DEST { "SESSION STATUS RESULT=DUPLICATED_DEST\n" };
+	constexpr std::string_view SAM_SESSION_CREATE_INVALID_ID { "SESSION STATUS RESULT=INVALID_ID\n" };
+	constexpr std::string_view SAM_SESSION_STATUS_INVALID_KEY { "SESSION STATUS RESULT=INVALID_KEY\n" };
 	const char SAM_SESSION_STATUS_I2P_ERROR[] = "SESSION STATUS RESULT=I2P_ERROR MESSAGE=\"%s\"\n";
 	const char SAM_SESSION_ADD[] = "SESSION ADD";
 	const char SAM_SESSION_REMOVE[] = "SESSION REMOVE";
 	const char SAM_STREAM_CONNECT[] = "STREAM CONNECT";
-	const char SAM_STREAM_STATUS_OK[] = "STREAM STATUS RESULT=OK\n";
-	const char SAM_STREAM_STATUS_INVALID_ID[] = "STREAM STATUS RESULT=INVALID_ID\n";
-	const char SAM_STREAM_STATUS_INVALID_KEY[] = "STREAM STATUS RESULT=INVALID_KEY\n";
+	constexpr std::string_view SAM_STREAM_STATUS_OK { "STREAM STATUS RESULT=OK\n" };
+	constexpr std::string_view SAM_STREAM_STATUS_INVALID_ID { "STREAM STATUS RESULT=INVALID_ID\n" };
+	constexpr std::string_view SAM_STREAM_STATUS_INVALID_KEY { "STREAM STATUS RESULT=INVALID_KEY\n" };
 	const char SAM_STREAM_STATUS_CANT_REACH_PEER[] = "STREAM STATUS RESULT=CANT_REACH_PEER MESSAGE=\"%s\"\n";
 	const char SAM_STREAM_STATUS_I2P_ERROR[] = "STREAM STATUS RESULT=I2P_ERROR MESSAGE=\"%s\"\n";
 	const char SAM_STREAM_ACCEPT[] = "STREAM ACCEPT";
@@ -64,7 +66,7 @@ namespace client
 	const char SAM_DEST_REPLY_I2P_ERROR[] = "DEST REPLY RESULT=I2P_ERROR\n";
 	const char SAM_NAMING_LOOKUP[] = "NAMING LOOKUP";
 	const char SAM_NAMING_REPLY[] = "NAMING REPLY RESULT=OK NAME=%s VALUE=%s\n";
-	const char SAM_DATAGRAM_RECEIVED[] = "DATAGRAM RECEIVED DESTINATION=%s SIZE=%lu\n";
+	const char SAM_DATAGRAM_RECEIVED[] = "DATAGRAM RECEIVED DESTINATION=%s SIZE=%lu FROM_PORT=%u TO_PORT=%u\n";
 	const char SAM_RAW_RECEIVED[] = "RAW RECEIVED SIZE=%lu\n";
 	const char SAM_NAMING_REPLY_INVALID_KEY[] = "NAMING REPLY RESULT=INVALID_KEY NAME=%s\n";
 	const char SAM_NAMING_REPLY_KEY_NOT_FOUND[] = "NAMING REPLY RESULT=KEY_NOT_FOUND NAME=%s\n";
@@ -81,15 +83,25 @@ namespace client
 	const char SAM_PARAM_HOST[] = "HOST";
 	const char SAM_PARAM_PORT[] = "PORT";
 	const char SAM_PARAM_FROM_PORT[] = "FROM_PORT";
+	const char SAM_PARAM_TO_PORT[] = "TO_PORT";
 	const char SAM_VALUE_TRANSIENT[] = "TRANSIENT";
-	const char SAM_VALUE_STREAM[] = "STREAM";
-	const char SAM_VALUE_DATAGRAM[] = "DATAGRAM";
-	const char SAM_VALUE_RAW[] = "RAW";
-	const char SAM_VALUE_MASTER[] = "MASTER";
 	const char SAM_VALUE_TRUE[] = "true";
 	const char SAM_VALUE_FALSE[] = "false";
 
-	enum SAMSocketType
+	constexpr std::string_view SAM_VALUE_STREAM { "STREAM" };
+	constexpr std::string_view SAM_VALUE_DATAGRAM { "DATAGRAM" };
+	constexpr std::string_view SAM_VALUE_RAW { "RAW" };
+	constexpr std::string_view SAM_VALUE_MASTER { "MASTER" };
+
+	constexpr std::string_view SAM_PING { "PING" };
+	const char SAM_PONG[] = "PONG %s\n";
+
+	constexpr int MAKE_SAM_VERSION_NUMBER (int major, int minor) { return major*10 + minor; }
+	constexpr int MIN_SAM_VERSION = MAKE_SAM_VERSION_NUMBER (3, 0);
+	constexpr int MAX_SAM_VERSION = MAKE_SAM_VERSION_NUMBER (3, 3);
+	constexpr int SAM_VERSION_33 = MAKE_SAM_VERSION_NUMBER (3, 3); // SAM 3.3
+
+	enum class SAMSocketType
 	{
 		eSAMSocketTypeUnknown,
 		eSAMSocketTypeSession,
@@ -116,7 +128,9 @@ namespace client
 
 			void Terminate (const char* reason);
 
-			bool IsSession(const std::string & id) const;
+			bool IsSession(std::string_view id) const;
+
+			static i2p::util::Mapping ExtractParams (std::string_view buf);
 
 		private:
 
@@ -125,7 +139,7 @@ namespace client
 			void HandleHandshakeReceived (const boost::system::error_code& ecode, std::size_t bytes_transferred);
 			void HandleHandshakeReplySent (const boost::system::error_code& ecode, std::size_t bytes_transferred);
 			void HandleMessage (const boost::system::error_code& ecode, std::size_t bytes_transferred);
-			void SendMessageReply (const char * msg, size_t len, bool close);
+			void SendMessageReply (std::string_view msg, bool close);
 			void HandleMessageReplySent (const boost::system::error_code& ecode, std::size_t bytes_transferred, bool close);
 			void Receive ();
 			void HandleReceived (const boost::system::error_code& ecode, std::size_t bytes_transferred);
@@ -135,23 +149,24 @@ namespace client
 			void HandleI2PAccept (std::shared_ptr<i2p::stream::Stream> stream);
 			void HandleI2PForward (std::shared_ptr<i2p::stream::Stream> stream, boost::asio::ip::tcp::endpoint ep);
 			void HandleWriteI2PData (const boost::system::error_code& ecode, size_t sz);
-			void HandleI2PDatagramReceive (const i2p::data::IdentityEx& from, uint16_t fromPort, uint16_t toPort, const uint8_t * buf, size_t len);
+			void HandleI2PDatagramReceive (const i2p::data::IdentityEx& from, uint16_t fromPort, uint16_t toPort,
+				const uint8_t * buf, size_t len, const i2p::util::Mapping * options);
 			void HandleI2PRawDatagramReceive (uint16_t fromPort, uint16_t toPort, const uint8_t * buf, size_t len);
 
-			void ProcessSessionCreate (char * buf, size_t len);
+			void ProcessSessionCreate (std::string_view buf);
 			void ProcessStreamConnect (char * buf, size_t len, size_t rem);
-			void ProcessStreamAccept (char * buf, size_t len);
-			void ProcessStreamForward (char * buf, size_t len);
-			void ProcessDestGenerate (char * buf, size_t len);
-			void ProcessNamingLookup (char * buf, size_t len);
-			void ProcessSessionAdd (char * buf, size_t len);
-			void ProcessSessionRemove (char * buf, size_t len);
+			void ProcessStreamAccept (std::string_view buf);
+			void ProcessStreamForward (std::string_view buf);
+			void ProcessDestGenerate (std::string_view buf);
+			void ProcessNamingLookup (std::string_view buf);
+			void ProcessSessionAdd (std::string_view buf);
+			void ProcessSessionRemove (std::string_view buf);
+			void ProcessPing (std::string_view text);
 			void SendReplyWithMessage (const char * reply, const std::string & msg);
 			void SendSessionI2PError(const std::string & msg);
-			void SendStreamI2PError(const std::string & msg);	
+			void SendStreamI2PError(const std::string & msg);
 			void SendStreamCantReachPeer(const std::string & msg);
 			size_t ProcessDatagramSend (char * buf, size_t len, const char * data); // from SAM 1.0
-			void ExtractParams (char * buf, std::map<std::string, std::string>& params);
 
 			void Connect (std::shared_ptr<const i2p::data::LeaseSet> remote, std::shared_ptr<SAMSession> session = nullptr);
 			void HandleConnectLeaseSetRequestComplete (std::shared_ptr<i2p::data::LeaseSet> leaseSet);
@@ -180,9 +195,10 @@ namespace client
 			bool m_IsAccepting; // for eSAMSocketTypeAcceptor only
 			bool m_IsReceiving; // for eSAMSocketTypeStream only
 			std::shared_ptr<i2p::stream::Stream> m_Stream;
+			int m_Version;
 	};
 
-	enum SAMSessionType
+	enum class SAMSessionType
 	{
 		eSAMSessionTypeUnknown,
 		eSAMSessionTypeStream,
@@ -196,10 +212,12 @@ namespace client
 		SAMBridge & m_Bridge;
 		std::string Name;
 		SAMSessionType Type;
+		i2p::datagram::DatagramVersion DatagramVersion;
+
 		std::shared_ptr<boost::asio::ip::udp::endpoint> UDPEndpoint; // TODO: move
 		std::list<std::pair<std::shared_ptr<SAMSocket>, uint64_t> > acceptQueue; // socket, receive time in seconds
-		
-		SAMSession (SAMBridge & parent, const std::string & name, SAMSessionType type);
+
+		SAMSession (SAMBridge & parent, std::string_view name, SAMSessionType type);
 		virtual ~SAMSession () {};
 
 		virtual std::shared_ptr<ClientDestination> GetLocalDestination () = 0;
@@ -213,7 +231,7 @@ namespace client
 	{
 		std::shared_ptr<ClientDestination> localDestination;
 
-		SAMSingleSession (SAMBridge & parent, const std::string & name, SAMSessionType type, std::shared_ptr<ClientDestination> dest);
+		SAMSingleSession (SAMBridge & parent, std::string_view name, SAMSessionType type, std::shared_ptr<ClientDestination> dest);
 		~SAMSingleSession ();
 
 		std::shared_ptr<ClientDestination> GetLocalDestination () { return localDestination; };
@@ -222,9 +240,9 @@ namespace client
 
 	struct SAMMasterSession: public SAMSingleSession
 	{
-		std::set<std::string> subsessions;
-		SAMMasterSession (SAMBridge & parent, const std::string & name, std::shared_ptr<ClientDestination> dest):
-			SAMSingleSession (parent, name, eSAMSessionTypeMaster, dest) {};
+		std::set<std::string, std::less<> > subsessions;
+		SAMMasterSession (SAMBridge & parent, std::string_view name, std::shared_ptr<ClientDestination> dest):
+			SAMSingleSession (parent, name, SAMSessionType::eSAMSessionTypeMaster, dest) {};
 		void Close ();
 	};
 
@@ -233,7 +251,7 @@ namespace client
 		std::shared_ptr<SAMMasterSession> masterSession;
 		uint16_t inPort;
 
-		SAMSubSession (std::shared_ptr<SAMMasterSession> master, const std::string& name, SAMSessionType type, uint16_t port);
+		SAMSubSession (std::shared_ptr<SAMMasterSession> master, std::string_view name, SAMSessionType type, uint16_t port);
 		// implements SAMSession
 		std::shared_ptr<ClientDestination> GetLocalDestination ();
 		void StopLocalDestination ();
@@ -250,13 +268,13 @@ namespace client
 			void Stop ();
 
 			auto& GetService () { return GetIOService (); };
-			std::shared_ptr<SAMSession> CreateSession (const std::string& id, SAMSessionType type, const std::string& destination, // empty string means transient
-				const std::map<std::string, std::string> * params);
+			std::shared_ptr<SAMSession> CreateSession (std::string_view id, SAMSessionType type, std::string_view destination, // empty string means transient
+				const i2p::util::Mapping& params);
 			bool AddSession (std::shared_ptr<SAMSession> session);
-			void CloseSession (const std::string& id);
-			std::shared_ptr<SAMSession> FindSession (const std::string& id) const;
+			void CloseSession (std::string_view id);
+			std::shared_ptr<SAMSession> FindSession (std::string_view id) const;
 
-			std::list<std::shared_ptr<SAMSocket> > ListSockets(const std::string & id) const;
+			std::list<std::shared_ptr<SAMSocket> > ListSockets(std::string_view id) const;
 
 			/** send raw data to remote endpoint from our UDP Socket */
 			void SendTo (const std::vector<boost::asio::const_buffer>& bufs, const boost::asio::ip::udp::endpoint& ep);
@@ -264,7 +282,7 @@ namespace client
 			void AddSocket(std::shared_ptr<SAMSocket> socket);
 			void RemoveSocket(const std::shared_ptr<SAMSocket> & socket);
 
-			bool ResolveSignatureType (const std::string& name, i2p::data::SigningKeyType& type) const;
+			bool ResolveSignatureType (std::string_view name, i2p::data::SigningKeyType& type) const;
 
 		private:
 
@@ -276,8 +294,8 @@ namespace client
 
 			void ScheduleSessionCleanupTimer (std::shared_ptr<SAMSession> session);
 			void HandleSessionCleanupTimer (const boost::system::error_code& ecode,
-				std::shared_ptr<SAMSession> session, std::shared_ptr<boost::asio::deadline_timer> timer);	                                
-			
+				std::shared_ptr<SAMSession> session, std::shared_ptr<boost::asio::deadline_timer> timer);
+
 		private:
 
 			bool m_IsSingleThread;
@@ -285,11 +303,11 @@ namespace client
 			boost::asio::ip::udp::endpoint m_DatagramEndpoint, m_SenderEndpoint;
 			boost::asio::ip::udp::socket m_DatagramSocket;
 			mutable std::mutex m_SessionsMutex;
-			std::map<std::string, std::shared_ptr<SAMSession> > m_Sessions;
+			std::map<std::string, std::shared_ptr<SAMSession>, std::less<>> m_Sessions;
 			mutable std::mutex m_OpenSocketsMutex;
 			std::list<std::shared_ptr<SAMSocket> > m_OpenSockets;
 			uint8_t m_DatagramReceiveBuffer[i2p::datagram::MAX_DATAGRAM_SIZE+1];
-			std::map<std::string, i2p::data::SigningKeyType> m_SignatureTypes;
+			const std::map<std::string_view, i2p::data::SigningKeyType> m_SignatureTypes;
 
 		public:
 
